module Bernstein


using dg1d
using LinearAlgebra
using Jacobi


# Normalized so that bernstein_polynomial(-1,0,N) == bernstein_polynomial(1,N,N) == 1
function bernstein_polynomial(x::Real, n::Integer, N::Integer)
  return binomial(N, n) / 2^N * (1 + x)^n * (1 - x)^(N-n)
end


function derivative_bernstein_polynomial(x::Real, n::Integer, N::Integer)
  B = 0.0
  if n > 0
    B += n*(1+x)^(n-1) * (1-x)^(N-n)
  end
  if n < N
    B += (1+x)^n * (N-n)*(1-x)^(N-n-1)*(-1)
  end
  return binomial(N, n) / 2^N * B
end


function vandermonde_matrix(z::AbstractVector)
  Npts = length(z)
  V = [ bernstein_polynomial(z[i], j-1, Npts-1) for i=1:Npts, j=1:Npts ]
  return V
end


"""
  mass_matrix(Npts::Integer)

Compute M_ij = (Bi, Bj) exactly where `Bi, Bj` are (i-1)th and (j-1)th Bernstein polynomials.
"""
function mass_matrix(Npts::Integer)
  N = Npts-1
  M = zeros(Float64, Npts, Npts)
  xs, ws = dg1d.LGL.rule(Npts+1)
  for i in 1:Npts, j in i:Npts
    mij = dg1d.LGL.integrate(xs, ws) do xx
      bi = bernstein_polynomial(xx, i-1, N)
      bj = bernstein_polynomial(xx, j-1, N)
      return bi*bj
    end
    M[i,j] = mij
    if j > i
      M[j,i] = mij
    end
  end
  return M
end


"""
  stiffness_matrix(Npts::Integer)

Compute S_ij = (Bi, dBj/dx) exactly where `Bi, Bj` are (i-1)th and (j-1)th Bernstein polynomials.
"""
function stiffness_matrix(Npts::Integer)
  N = Npts-1
  S = zeros(Float64, Npts, Npts)
  xs, ws = dg1d.LGL.rule(Npts+1)
  for i in 1:Npts, j in 1:Npts
    sij = dg1d.LGL.integrate(xs, ws) do xx
      bi = bernstein_polynomial(xx, i-1, N)
      ∂bj = derivative_bernstein_polynomial(xx, j-1, N)
      return bi*∂bj
    end
    S[i,j] = sij
  end
  return S
end


"""
  bernstein_legendre_transformation(Npts::Integer) -> (T, invT)

Compute a transformation matrix `T` and its inverse `invT` to map
quasi-nodal Bernstein coefficients `Bn` to modal Legendre coefficients `Ln`
by `Bn = T*Ln`.
"""
function bernstein_legendre_transformation(Npts::Integer)
  N = Npts-1
  xs, ws = dg1d.LGL.rule(Npts+1)
  BP = (i,j) -> dg1d.LGL.integrate(xs, ws) do xi
    Bim1 = bernstein_polynomial(xi, i-1, N)
    Pjm1 = Jacobi.legendre(xi, j-1)
    Bim1*Pjm1
  end
  PP = i -> dg1d.LGL.integrate(xs, ws) do xi
    Pim1 = Jacobi.legendre(xi, i-1)
    Pim1*Pim1
  end

  T = zeros(Float64, Npts, Npts)
  for i in 1:Npts, j in 1:Npts
    T[i,j] = BP(i,j)/PP(j)
  end
  invT = inv(T)

  return T, invT
end


"""
  interpolate(x::Real, fn::AbstractVector)

Interpolate a Bernstein polynomial with quasi-nodal coefficients `fn` onto `x`.
`x` is assumed to lie within `[-1,1`].
The `fn` are assumed to be sampled on the nodes `xs = [ -1.0 + 2*i/N for i in 0:N ]`
where `N = length(fn)-1`.
"""
function interpolate(x::Real, fn::AbstractVector)
  Npts = length(fn)
  N = Npts-1
  return sum(enumerate(fn)) do (i,fi)
    Bim1 = bernstein_polynomial(x, i-1, N)
    fi*Bim1
  end
end


end
