@testset "Spectral Element" begin


function evaluate_derivatives(el::DGElement, f, df)
  f_ana = f.(el.z)
  df_ana = df.(el.z)
  df_num = el.D * f_ana
  return df_ana, df_num
end


@testset "Nodal derivatives" begin

  N = 4
  f(x) = x^3 - 5.0 * x^2 + 1.0
  df(x) = 3.0 * x^2 - 10.0 * x
  # ddf(x) = 6.0 * x - 10.0

  dg_kind = :nodal_lagrange

  el = DGElement(N, :lgl, dg_kind)
  df_ana, df_num = evaluate_derivatives(el, f, df)
  @test isapprox(df_ana, df_num)

  el = DGElement(N, :lgl_lumped, dg_kind)
  df_ana, df_num = evaluate_derivatives(el, f, df)
  @test isapprox(df_ana, df_num)

  glgl = DGElement(N, :lgl, dg_kind)
  df_ana, df_num = evaluate_derivatives(el, f, df)
  @test isapprox(df_ana, df_num)

end


@testset "Quadrature check" begin

  # integrate f^2 over x ∈ [-1,1]
  f(x) = x^3 - 5.0 * x^2 + 1.0
  f2_ana = 118.0/21.0 # according to wolfram alpha

  # LGL quadrature is only exact up to order 2*3-1=5, so use N=4 instead
  N = 4
  L = 2.0
  quadr = :lgl
  el = DGElement(N, quadr, :nodal_lagrange)
  fx = f.(el.z)
  f2 = dg1d.inner_product(fx, fx, el)
  @test isapprox(f2, f2_ana)

  # check that :glgl quadrature is exact up to and including order 2N
  N = 3
  L = 2.0
  el = DGElement(N, :glgl, :nodal_lagrange)
  fx = f.(el.z)
  f2 = dg1d.inner_product(fx, fx, el)
  @test isapprox(f2, f2_ana)

end


@testset "Barycentric interpolation" begin

  # setup
  Npts = 6
  x, _ = dg1d.LGL.rule(Npts)
  bw = dg1d.barycentric_weights(x)

  # test function
  f(x) = x
  # f(x) = x^4 - 2*x + 1
  fx = @. f(x)

  # test interpolation on random grid
  nsmpl = 3
  xsmpl = rand(nsmpl) # ∈ [0,1]
  @. xsmpl = 2 * xsmpl - 1 # ∈ [-1,1]
  fxsmpl = @. f(xsmpl)

  # test single point interpolation
  fxintrp = [ dg1d.barycentric_interp(xi, x, bw, fx) for xi in xsmpl ]
  @test all(isapprox.(fxintrp, fxsmpl))

  # test interpolation using matrix multiplication
  I = dg1d.barycentric_interp_matrix(xsmpl, x)
  fxintrp = I * fx
  @test all(isapprox.(fxintrp, fxsmpl))

  # edge case: interpolate onto nodes themselves
  xsmpl = x
  fxsmpl = fx
  fxintrp = [ dg1d.barycentric_interp(xi, x, bw, fx) for xi in xsmpl ]
  @test all(isapprox.(fxintrp, fxsmpl))

  I = dg1d.barycentric_interp_matrix(xsmpl, x)
  fxintrp = I * fx
  @test all(isapprox.(fxintrp, fxsmpl))

end


@testset "Test orthogonality relation GLGL" begin

  Npts = 6

  x, w, v, D = dg1d.GLGL.inner_product_rule(Npts)
  L = zeros(Npts,Npts)
  for i in 1:Npts
    Pi = map(x->dg1d.legendre(x, i-1), x)
    for j in i:Npts
      Pj = map(x->dg1d.legendre(x, j-1), x)
      lij = dg1d.GLGL.inner_product(Pi, Pj, x, w, v, D)
      L[i,j] = lij
      if i != j
        L[j,i] = L[i,j]
      end
    end
  end
  dg1d.purge_zeros!(L)

  # analytical result
  L_ana = zeros(Npts, Npts)
  for i in 1:Npts
    L_ana[i,i] = 2.0 / (2.0 * i - 1.0)
  end

  @test isapprox(L, L_ana)

end


function lgl_linear_dependency_matrix_exact(Npts::Integer)
  Npts2 = Npts+2
  x2, w2 = dg1d.LGL.rule(Npts2)
  L = zeros(N,N)
  for i in 1:N
    Pi = map(x->dg1d.legendre(x, i-1), x2)
    for j in i:N
      Pj = map(x->dg1d.legendre(x, j-1), x2)
      lij = dg1d.LGL.integrate(Pi, Pj, x2, w2)
      L[i,j] = lij
      if i != j
        L[j,i] = L[i,j]
      end
    end
  end
  dg1d.purge_zeros!(L)

  return L
end


@testset "Test orthogonality relation" begin

  Npts = 6

  Npts2 = Npts+2
  x2, w2 = dg1d.LGL.rule(Npts2)
  L = zeros(Npts,Npts)
  for i in 1:Npts
    Pi = map(x->dg1d.legendre(x, i-1), x2)
    for j in i:Npts
      Pj = map(x->dg1d.legendre(x, j-1), x2)
      lij = dg1d.LGL.inner_product(Pi, Pj, x2, w2)
      L[i,j] = lij
      if i != j
        L[j,i] = L[i,j]
      end
    end
  end
  dg1d.purge_zeros!(L)

  # analytical result
  L_ana = zeros(Npts, Npts)
  for i in 1:Npts
    L_ana[i,i] = 2.0 / (2.0 * i - 1.0)
  end

  @test isapprox(L, L_ana)

end


@testset "More integrator tests" begin

  Npts = 6
  p = x -> (x-4)^2 * (x+1)^3

  x, w = dg1d.LGL.rule(Npts)
  int = dg1d.LGL.integrate(p, x, w)
  @test int ≈ 140/3

  x, w = dg1d.GLGL.rule(Npts)
  int = dg1d.GLGL.integrate(p, x, w)
  @test int ≈ 140/3

end


@testset "verify some DG operator identities" begin

  N = 10

  for basis in (:lgl, :glgl)
    el = dg1d.DGElement(N, basis, :nodal_lagrange)
    # verify: M D = S
    MD = dg1d.purge_zeros(el.M*el.D)
    S = dg1d.purge_zeros(el.S)
    @test all(MD.≈S)
  end

  el = dg1d.DGElement(N, :lgl, :modal_bernstein)
  T, invT = dg1d.Bernstein.bernstein_legendre_transformation(el.Npts)
  M_legendre = dg1d.Legendre.mass_matrix(el.Npts)
  TMT = T*M_legendre*transpose(T) |> dg1d.purge_zeros
  @test all(el.M.≈TMT)
  S_legendre = dg1d.Legendre.stiffness_matrix(el.Npts)
  TST = T*S_legendre*transpose(T) |> dg1d.purge_zeros
  @test all(el.S.≈TST)

end


@testset "Legendre expansions and Bernstein polynomials" begin

  # approximate the Bernstein basis through Legendre polynomials
  N = 10
  Npts = N+1
  Cs = zeros(Float64, Npts, Npts)
  for i in 1:Npts, j in 1:Npts
    # ith Bernstein polynomial, jth Legendre modal coefficient
    Cs[i,:] .= dg1d.Legendre.projection(Npts) do xi
      dg1d.Bernstein.bernstein_polynomial(xi, i-1, N)
    end
  end

  xx = collect(range(-1.0,1.0,25))
  for xi in xx
    Bs = [ dg1d.Bernstein.bernstein_polynomial(xi, i-1, N) for i in 1:Npts ] |> dg1d.purge_zeros
    approx_Bs = [ dg1d.Legendre.interpolate(xi, Cs[i,:]) for i in 1:Npts ] |> dg1d.purge_zeros
    @test all(isapprox.(Bs, approx_Bs; atol=1e-10, rtol=1e-8))
  end

end


end
